/*
 * JBoss, Home of Professional Open Source.
 * Copyright 2008, Red Hat Middleware LLC, and individual contributors
 * as indicated by the @author tags. See the copyright.txt file in the
 * distribution for a full listing of individual contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */
package org.jboss.web.tomcat.service.session.distributedcache.impl.jbc;

import java.util.Map;

import org.jboss.cache.Fqn;
import org.jboss.cache.notifications.annotation.NodeModified;
import org.jboss.cache.notifications.annotation.NodeRemoved;
import org.jboss.cache.notifications.event.NodeModifiedEvent;
import org.jboss.cache.notifications.event.NodeRemovedEvent;
import org.jboss.logging.Logger;
import org.jboss.metadata.web.jboss.ReplicationGranularity;
import org.jboss.web.tomcat.service.session.distributedcache.spi.DistributableSessionMetadata;
import org.jboss.web.tomcat.service.session.distributedcache.spi.LocalDistributableSessionManager;

/**
 * Listens for removals and modifications in the cache, notifying the 
 * session manager of significant events.
 * 
 * @author Brian Stansberry
 */
@org.jboss.cache.notifications.annotation.CacheListener
public class CacheListener extends CacheListenerBase
{
   // Element within an FQN that is the root of a Pojo attribute map
   private static final int POJO_ATTRIBUTE_FQN_INDEX = SESSION_ID_FQN_INDEX + 1;
   // Element within an FQN that is the root of an individual Pojo attribute
   private static final int POJO_KEY_FQN_INDEX = POJO_ATTRIBUTE_FQN_INDEX + 1;
   // Element within an FQN that is the root of a session's internal pojo storage area
   private static final int POJO_INTERNAL_FQN_INDEX = SESSION_ID_FQN_INDEX + 1;
   // Minimum size of an FQN that is below the root of a session's internal pojo storage area
   private static final int POJO_INTERNAL_FQN_SIZE = POJO_INTERNAL_FQN_INDEX + 1;
   private static Logger log_ = Logger.getLogger(CacheListener.class);
   private boolean attributeBased_;
   private boolean fieldBased_;

   CacheListener(JBossCacheWrapper wrapper, LocalDistributableSessionManager manager, String contextHostPath, ReplicationGranularity granularity)
   {      
      super(manager, contextHostPath);
      if (granularity == ReplicationGranularity.FIELD)
         fieldBased_ = true;
      else if (granularity == ReplicationGranularity.ATTRIBUTE)
         attributeBased_ = true;
   }

   // --------------- CacheListener methods ------------------------------------

   @NodeRemoved
   public void nodeRemoved(NodeRemovedEvent event)
   {      
      if (event.isPre())
         return;
      
      boolean local = event.isOriginLocal();
      if (!fieldBased_ && local)
         return;
      
      @SuppressWarnings("unchecked")
      Fqn<String> fqn = event.getFqn();
      boolean isBuddy = isBuddyFqn(fqn); 
      int baseIdx = Util.getJSessionIndex(fqn, isBuddy);
      
      if (!local 
            && isFqnSessionRootSized(fqn, baseIdx) 
            && isFqnForOurWebapp(fqn, baseIdx))
      {
         // A session has been invalidated from another node;
         // need to inform manager
         String sessId = getIdFromFqn(fqn, baseIdx);
         manager_.notifyRemoteInvalidation(sessId);
      }
      else if (local && !isBuddy
                  && isPossibleInternalPojoFqn(fqn, baseIdx) 
                  && isFqnForOurWebapp(fqn, baseIdx))
      {
         // One of our sessions' pojos is modified; need to inform
         // the manager so it can mark the session dirty
         String sessId = getIdFromFqn(fqn, baseIdx);
         manager_.notifyLocalAttributeModification(sessId);
      }
   }   
   
   
   @NodeModified
   public void nodeModified(NodeModifiedEvent event)
   {      
      if (event.isPre())
         return;
      
      boolean local = event.isOriginLocal();
      if (!fieldBased_ && local)
         return;
      
      @SuppressWarnings("unchecked")
      Fqn<String> fqn = event.getFqn();
      boolean isBuddy = isBuddyFqn(fqn);  
      int baseIdx = Util.getJSessionIndex(fqn, isBuddy);
      
      if (!local 
             && isFqnSessionRootSized(fqn, baseIdx)
             && isFqnForOurWebapp(fqn, baseIdx))
      {
         // Query if we have version value in the distributed cache. 
         // If we have a version value, compare the version and invalidate if necessary.
         @SuppressWarnings("unchecked")
         Map<Object, Object> data = event.getData();
         Integer version = (Integer) data.get(AbstractJBossCacheService.VERSION_KEY);
         if(version != null)
         {
            String realId = getIdFromFqn(fqn, baseIdx);
            String owner = isBuddy ? Util.getBuddyOwner(fqn) : null;
            Long timestamp = (Long) data.get(AbstractJBossCacheService.TIMESTAMP_KEY);
            if (timestamp == null)
            {
               log_.warn("No timestamp attribute found in node modification " +
               		"event data for session " + Util.maskId(realId) + 
               		" under parent fqn " + fqn.getParent());
            }
            else
            {
               // Notify the manager that a session has been updated
               boolean updated = manager_.sessionChangedInDistributedCache(realId, owner, 
                                                  version.intValue(), 
                                                  timestamp.longValue(), 
                                                  (DistributableSessionMetadata) data.get(AbstractJBossCacheService.METADATA_KEY));
               if (!updated && !isBuddy)
               {
                  log_.warn("Possible concurrency problem: Replicated version id " + 
                            version + " is less than or equal to in-memory version for session " + realId); 
               }
               /*else 
               {
                  We have a local session but got a modification for the buddy tree.
                  This means another node is in the process of taking over the session;
                  we don't worry about it
               }
                */
            }
         }
         else if (!attributeBased_) // other granularities can modify attributes only
         {
            log_.warn("No version attribute found in " + fqn);
         }
      }
      else if (local && !isBuddy
            && isPossibleInternalPojoFqn(fqn, baseIdx) 
            && isFqnForOurWebapp(fqn, baseIdx))
      {
         // One of our sessions' pojos is modified; need to inform
         // the manager so it can mark the session dirty
         String sessId = getIdFromFqn(fqn, baseIdx);
         manager_.notifyLocalAttributeModification(sessId);
      }
   }
   
   public static String getPojoKeyFromFqn(Fqn<String> fqn, int jsessionIndex)
   {
      return (String) fqn.get(jsessionIndex + POJO_KEY_FQN_INDEX);
   }

   @Deprecated
   public static String getPojoKeyFromFqn(Fqn<String> fqn, boolean isBuddy)
   {
      int index = Util.getJSessionIndex(fqn, isBuddy);
      return getPojoKeyFromFqn(fqn, index);
   }
   
   /**
    * Check if the fqn is big enough to be in the internal pojo area but
    * isn't in the regular attribute area.
    * 
    * Structure in the cache is:
    * 
    * /JSESSION
    * ++ /contextPath_hostname
    * ++++ /sessionid
    * ++++++ /ATTRIBUTE
    * ++++++ /_JBossInternal_
    * ++++++++ etc etc
    * 
    * If the Fqn size is big enough to be "etc etc" or lower, but the 4th
    * level is not "ATTRIBUTE", it must be under _JBossInternal_. We discriminate
    * based on != ATTRIBUTE to avoid having to code to the internal PojoCache
    * _JBossInternal_ name.
    * 
    * @param fqn the 
    * @param
    * @return
    */
   public static boolean isPossibleInternalPojoFqn(Fqn<String> fqn, int jsessionIndex)
   {      
      return (jsessionIndex > -1 && fqn.size() > jsessionIndex + POJO_INTERNAL_FQN_SIZE 
            && FieldBasedJBossCacheService.ATTRIBUTE.equals(fqn.get(jsessionIndex + POJO_INTERNAL_FQN_INDEX)) == false);
   }
   
   /**
    * Check if the fqn is big enough to be in the internal pojo area but
    * isn't in the regular attribute area.
    * 
    * Structure in the cache is:
    * 
    * /JSESSION
    * ++ /contextPath_hostname
    * ++++ /sessionid
    * ++++++ /ATTRIBUTE
    * ++++++ /_JBossInternal_
    * ++++++++ etc etc
    * 
    * If the Fqn size is big enough to be "etc etc" or lower, but the 4th
    * level is not "ATTRIBUTE", it must be under _JBossInternal_. We discriminate
    * based on != ATTRIBUTE to avoid having to code to the internal PojoCache
    * _JBossInternal_ name.
    * 
    * @param fqn
    * @return
    */
   @Deprecated
   public static boolean isPossibleInternalPojoFqn(Fqn<String> fqn)
   {      
      int index = Util.getJSessionIndex(fqn, false);
      return isPossibleInternalPojoFqn(fqn, index);
   }
}
